{-# OPTIONS --rewriting #-}
open import Data.Nat as ℕ
open import Data.Nat.Properties as ℕ
open import Data.Fin as Fin using (Fin; zero; suc; remQuot; combine; quotRem)
open import Data.Fin.Properties as Fin using (remQuot-combine; combine-remQuot)
open import Relation.Binary.PropositionalEquality as Eq hiding ([_])
open import Relation.Binary
open import Data.Product as Prod hiding (map)
open import Data.Sum hiding (map; swap)
open import Function
open import Algebra
open import Level using () renaming (zero to ℓ₀)

{-# BUILTIN REWRITE _≡_ #-}

variable
  X Y Z : Set
  m n k : ℕ

infixl 5 _⊗_

data S : Set where
  ι    : ℕ → S
  _⊗_  : S → S → S

variable
  s p q r u v w : S

data P : S → Set where
  ι    : Fin n → P (ι n)
  _⊗_  : P s → P p → P (s ⊗ p)

_⟪_⟫ : Set → S → Set
X ⟪ s ⟫ = P s → X

_[_]  : ∀ {s} → X ⟪ s ⟫ → (P s → X)
a [ ix ] = a ix

ιsuc : P (ι n) → P (ι (suc n))
ιsuc (ι i) = ι (suc i)

hd : X ⟪ ι (suc n) ⟫ → X
hd a = a [ ι zero ]

tl : X ⟪ ι (suc n) ⟫ → X ⟪ ι n ⟫
tl a i = a [ ιsuc i ]

sum₁ : (X → X → X) → X → X ⟪ ι n ⟫ → X
sum₁ {n = zero}   _⊕_ e a = e
sum₁ {n = suc n}  _⊕_ e a = hd a ⊕ sum₁ _⊕_ e (tl a)

K : X → X ⟪ s ⟫
K x i = x

map : (X → Y) → X ⟪ s ⟫ → Y ⟪ s ⟫
map f a i = f (a [ i ])

zipWith : (X → Y → Z) → X ⟪ s ⟫ → Y ⟪ s ⟫ → Z ⟪ s ⟫
zipWith f a b i = f (a [ i ]) (b [ i ])

nest : X ⟪ s ⊗ p ⟫ → X ⟪ p ⟫ ⟪ s ⟫
nest a i j = a [ i ⊗ j ]

unnest : X ⟪ p ⟫ ⟪ s ⟫ → X ⟪ s ⊗ p ⟫
unnest a (i ⊗ j) = a [ i ] [ j ]

sum : (X → X → X) → X → X ⟪ s ⟫ → X
sum {s = ι x}    f e a = sum₁  f e a
sum {s = s ⊗ p}  f e a = sum   f e $ map (sum f e) $ nest a

module MMnaive (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z) 
               (let infixl 10 _⊠_; _⊠_ = _⊠_) where

  -- Canonical matrix multiply without any blocking.
  mm-canon : X ⟪ s ⊗ p ⟫ → Y ⟪ p ⊗ q ⟫ → Z ⟪ s ⊗ q ⟫
  mm-canon a b (i ⊗ j) =
    sum _⊞_ ε λ k → a [ i ⊗ k ] ⊠ b [ k ⊗ j ]

_≈ₐ_ : X ⟪ s ⟫ → X ⟪ s ⟫ → Set
a ≈ₐ b = ∀ i → a [ i ] ≡ b [ i ]

-- Block over rows and columns
tile : X ⟪ (s ⊗ p) ⊗ (q ⊗ r) ⟫ → X ⟪ (s ⊗ q) ⊗ (p ⊗ r) ⟫
tile a ((i ⊗ j) ⊗ (k ⊗ l)) = a [ (i ⊗ k) ⊗ (j ⊗ l) ]

untile : X ⟪ (s ⊗ q) ⊗ (p ⊗ r) ⟫ → X ⟪ (s ⊗ p) ⊗ (q ⊗ r) ⟫
--untile a ((i ⊗ j) ⊗ (k ⊗ l)) = a [ (i ⊗ k) ⊗ (j ⊗ l) ]
untile = tile --a ((i ⊗ j) ⊗ (k ⊗ l)) = a [ (i ⊗ k) ⊗ (j ⊗ l) ]

--postulate
--    tile∘tile : (x : X ⟪ (s ⊗ p) ⊗ (q ⊗ r) ⟫) → tile (tile x) ≡ x
--
--{-# REWRITE tile∘tile #-}

--postulate
--  funext : ∀ {X Y : Set}{f g : X → Y} → (∀ x → f x ≡ g x) → f ≡ g 

-- Block over columns (make vertical cuts)
tileᵣ : X ⟪ s ⊗ (p ⊗ q) ⟫ → X ⟪ p ⊗ (s ⊗ q) ⟫
tileᵣ a (i ⊗ (j ⊗ k)) = a [ j ⊗ (i ⊗ k) ]

untileᵣ : X ⟪ p ⊗ (s ⊗ q) ⟫ → X ⟪ s ⊗ (p ⊗ q) ⟫
untileᵣ = tileᵣ --a (i ⊗ (j ⊗ k)) = a [ j ⊗ (i ⊗ k) ]

--- Tile over rows (make horizontal cuts)
tileₗ : X ⟪ (s ⊗ p) ⊗ q ⟫ → X ⟪ s ⊗ (p ⊗ q) ⟫
tileₗ a (i ⊗ (j ⊗ k)) = a [ (i ⊗ j) ⊗ k ]

untileₗ : X ⟪ s ⊗ (p ⊗ q) ⟫ → X ⟪ (s ⊗ p) ⊗ q ⟫
untileₗ a ((i ⊗ j) ⊗ k) = a [ i ⊗ (j ⊗ k) ]

refl-≈ₐ : {a : X ⟪ s ⟫} → a ≈ₐ a
refl-≈ₐ ix = refl

hd-cong : ∀ {n}{a b : X ⟪ ι (suc n) ⟫}
        → a ≈ₐ b → hd a ≡ hd b
hd-cong a≈b = a≈b (ι zero)

tl-cong : ∀ {n}{a b : X ⟪ ι (suc n) ⟫}
        → a ≈ₐ b → tl a ≈ₐ tl b
tl-cong a≈b i = a≈b (ιsuc i)

tl-comm : ∀ {n}{a : X ⟪ p ⟫ ⟪ ι (suc n) ⟫}
        → ∀ i → tl (λ j → a j i) ≡ λ j → tl a j i
tl-comm {n = n} i = refl

sum₁-comm : ∀ {n}{a : X ⟪ p ⟫ ⟪ ι n ⟫ }{f : X → X → X}{e : X}
          → sum₁ (zipWith f) (λ _ → e) a
            ≈ₐ λ i → sum₁ f e λ j → a j i
sum₁-comm {n = zero}  {a} {f} i = refl
sum₁-comm {n = suc n} {a} {f} i = cong₂ f refl (sum₁-comm {a = tl a} i)


sum₁-cong : ∀ {n}{a b : X ⟪ ι n ⟫}{f : X → X → X}{e : X}
          → a ≈ₐ b → sum₁ f e a ≡ sum₁ f e b
sum₁-cong {n = zero} a≈b = refl
sum₁-cong {n = suc n}{a = a}{b}{f}{e} a≈b = cong₂ f (hd-cong a≈b) (sum₁-cong $ tl-cong a≈b)

sum-cong : ∀ {s}{a b : X ⟪ s ⟫}{f : X → X → X}{e : X}
         → a ≈ₐ b → sum f e a ≡ sum f e b
sum-cong {s = ι x} a≈b = sum₁-cong a≈b
sum-cong {s = s ⊗ p} a≈b = sum-cong {s = s} λ i → sum-cong {s = p} λ j → a≈b _

sum-comm′ : {a : X ⟪ s ⊗ p ⟫}{f : X → X → X}{e : X}
          → sum (zipWith f) (λ _ → e) (nest a)
            ≈ₐ λ i → sum f e λ j → a (j ⊗ i)
sum-comm′ {s = ι x} i = sum₁-comm {n = x} i
sum-comm′ {s = s ⊗ p}{p = p₁}{a = a}{f}{e} i =
  let
    a′ = unnest (λ x →
              sum {s = p} (zipWith f) (λ _ → e)
              (λ j j₁ → a ((x ⊗ j) ⊗ j₁)))
    t = sum-comm′ {a = a′}{f}{e} i
  in trans t (sum-cong {s = s} λ ix → 
       let r = sum-comm′ { a = unnest (λ j j₁ → a (ix ⊗ j ⊗ j₁)) }{f}{e} i in r)


sum-comm : {a : X ⟪ p ⟫ ⟪ s ⟫}{f : X → X → X}{e : X}
          → sum (zipWith f) (K e) a
            ≈ₐ λ i → sum f e λ j → a [ j ] [ i ]
sum-comm {a = a} = sum-comm′ {a = unnest a}


fst : P (s ⊗ p) → P s
fst (i ⊗ _) = i
snd : P (s ⊗ p) → P p
snd (_ ⊗ j) = j
_⊛_ : P s → P p → P (s ⊗ p)
i ⊛ j = i ⊗ j
ψ : P (ι n) → Fin n
ψ (ι i) = i
φ : Fin n → P (ι n)
φ = ι

data Reshape : S → S → Set where
  eq     : Reshape s s
  _∙_    : Reshape p q → Reshape s p → Reshape s q
  _⊕_    : Reshape s p → Reshape q r → Reshape (s ⊗ q) (p ⊗ r)
  split  : Reshape (ι (m * n)) (ι m ⊗ ι n)
  flat   : Reshape (ι m ⊗ ι n) (ι (m * n))
  rtile   : Reshape ((s ⊗ p) ⊗ (q ⊗ r)) ((s ⊗ q) ⊗ (p ⊗ r))

_⟨_⟩ : ∀ {s p} → P p → Reshape s p → P s
i ⟨ eq ⟩ = i
i ⟨ r ⊕ r₁ ⟩ = (fst i ⟨ r ⟩) ⊛ (snd i ⟨ r₁ ⟩)
i ⟨ r ∙ r₁ ⟩ = i ⟨ r ⟩ ⟨ r₁ ⟩
i ⟨ split ⟩ = φ $ combine (ψ $ fst i) (ψ $ snd i)
i ⟨ flat ⟩ = let a , b = remQuot _ (ψ i) in φ a ⊛ φ b
(i ⊗ j ⊗ (k ⊗ l)) ⟨ rtile ⟩ = i ⊗ k ⊗ (j ⊗ l)

reshape : Reshape s p → X ⟪ s ⟫ →  X ⟪ p ⟫
reshape r a = λ ix → a [ ix ⟨ r ⟩ ]

rev : Reshape s p → Reshape p s
rev eq = eq
rev (r ⊕ r₁) = rev r ⊕ rev r₁
rev (r ∙ r₁) = rev r₁ ∙ rev r
rev split = flat
rev flat = split
rev rtile = rtile

rtile∘rtile : (i : P (s ⊗ p ⊗ (r ⊗ q))) → i ⟨ rtile ⟩ ⟨ rtile ⟩ ≡ i
rtile∘rtile (i ⊗ j ⊗ (k ⊗ l)) = refl

{-# REWRITE rtile∘rtile #-}


module MMblocked (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z)
         (let infixl 10 _⊠_; _⊠_ = _⊠_) 
         (let infixl 5 _⊞_; _⊞_ = _⊞_) where

  open MMnaive _⊠_ ε _⊞_
  ni : (A : Set) → A → A
  ni A a = a

  infix -2 ni
  syntax ni A e = e ∶ A

  -- Blocking dependeing on the ``form'' of the shape.
  mm : X ⟪ s ⊗ p ⟫ → Y ⟪ p ⊗ q ⟫ → Z ⟪ s ⊗ q ⟫

  -- All scalar dimensions in the shape, no blcocking.
  -- The base case for the induction.
  mm {s = ι m} {ι n} {ι k} = mm-canon

  mm {s = ι m} {ι n} {p ⊗ q} a b = 
    let  b′ = nest $ tileᵣ b                   ∶ Y ⟪ ι n ⊗ q ⟫ ⟪ p ⟫
         c′ = map (mm a) b′                    ∶ Z ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
    in   untileᵣ $ unnest c′

  mm {s = ι m} {p ⊗ q} {ι n} a b =
    let  a′ = nest $ tileᵣ a                   ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
         b′ = nest $ tileₗ b                   ∶ Y ⟪ q ⊗ ι n ⟫ ⟪ p ⟫
    in   sum (zipWith _⊞_) (K ε) λ k → mm (a′ [ k ]) (b′ [ k ])

  mm {s = ι m} {p ⊗ q} {r ⊗ w} a b =
    let  a′ = nest $ tileᵣ a                   ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
         b′ = nest $ tile b                    ∶ Y ⟪ q ⊗ w ⟫ ⟪ p ⊗ r ⟫
         c′ = λ (i : P r) → sum (zipWith _⊞_) (K ε) 
                                λ k → mm (a′ [ k ]) (b′ [ k ⊗ i ])
    in   untileᵣ $ unnest c′


  mm {s = s ⊗ p} {ι m} {ι n} a b =
    let  a′ = nest $ tileₗ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         c′ = map (flip mm b) a′               ∶ Z ⟪ p ⊗ ι n ⟫ ⟪ s ⟫
    in   untileₗ $ unnest c′

  mm {s = s ⊗ p}  {ι m} {q ⊗ r} a b =
    let  a′ = nest $ tileₗ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         b′ = nest $ tileᵣ b                   ∶ Y ⟪ ι m ⊗ r ⟫ ⟪ q ⟫
         c′ = (λ { (i ⊗ j) → mm (a′ [ i ]) 
                                (b′ [ j ]) })  ∶ Z ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
    in   untile $ unnest c′


  mm {s = s ⊗ p} {q ⊗ r} {ι m} a b =
    let  a′ = nest $ tile a                    ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ tileₗ b                   ∶ Y ⟪ r ⊗ ι m ⟫ ⟪ q ⟫
         c′ = λ (i : P s) → sum (zipWith _⊞_) (K ε) 
                                λ k → mm (a′ [ i ⊗ k ]) (b′ [ k ])
    in   untileₗ $ unnest c′

  -- Ok, here we reimplemented tile as "reshape rtile" which is
  -- exactly the same thing, except it has better normalisation properties.
  -- We should have eliminated tile entirely, but the point of this
  -- file is to demosntrate that we can do pre-tiling and we obtain the
  -- same results.
  mm {s = s ⊗ p} {q ⊗ r} {u ⊗ w} a b =
    let  a′ = nest $ reshape rtile a           ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ reshape rtile b           ∶ Y ⟪ r ⊗ w ⟫ ⟪ q ⊗ u ⟫
         c′ = λ { (i ⊗ j) → sum (zipWith _⊞_) (K ε) 
                            λ k → mm (a′ [ i ⊗ k ]) 
                                     (b′ [ k ⊗ j ]) }
    in   untile $ unnest c′



  mmt-canon : X ⟪ s ⊗ p ⟫ → Y ⟪ q ⊗ p ⟫ → Z ⟪ s ⊗ q ⟫
  mmt-canon a b (i ⊗ j) = sum _⊞_ ε λ k → a [ i ⊗ k ] ⊠ b [ j ⊗ k ]

  mmt : X ⟪ s ⊗ p ⟫ → Y ⟪ q ⊗ p ⟫ → Z ⟪ s ⊗ q ⟫
  mmt {s = ι m} {ι n} {ι k} = mmt-canon
  mmt {s = ι m} {ι n} {p ⊗ q} a b = let
      b′ = nest $ tileₗ b            ∶ Y ⟪ q ⊗ ι n ⟫ ⟪ p ⟫
      c′ = map (mmt a) b′            ∶ Z ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
      cc = untileᵣ $ unnest c′
    in cc

  mmt {s = ι m} {p ⊗ q} {ι n} a b = let
      a′ = nest $ tileᵣ a            ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
      b′ = nest $ tileᵣ b            ∶ Y ⟪ ι n ⊗ q ⟫ ⟪ p ⟫
    in sum (zipWith _⊞_) (K ε) λ k → mmt (a′ [ k ]) (b′ [ k ])

  mmt {s = ι m} {p ⊗ q} {r ⊗ w} a b = 
    let  a′ = nest $ tileᵣ a                   ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
         b′ = nest $ tile b                    ∶ Y ⟪ w ⊗ q ⟫ ⟪ r ⊗ p ⟫
         c′ = λ (i : P r) → sum (zipWith _⊞_) (K ε) 
                                λ k → mmt (a′ [ k ]) (b′ [ i ⊗ k ])
    in untileᵣ $ unnest c′

  mmt {s = s ⊗ p} {ι m} {ι n} a b =
    let  a′ = nest $ tileₗ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         c′ = map (flip mmt b) a′              ∶ Z ⟪ p ⊗ ι n ⟫ ⟪ s ⟫
    in untileₗ $ unnest c′

  mmt {s = s ⊗ p} {ι m} {q ⊗ r} a b = 
    let  a′ = nest $ tileₗ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         b′ = nest $ tileₗ b                   ∶ Y ⟪ r ⊗ ι m ⟫ ⟪ q ⟫
         c′ = (λ { (i ⊗ j) → mmt (a′ [ i ]) 
                                 (b′ [ j ]) })  ∶ Z ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
    in  untile $ unnest c′

  mmt {s = s ⊗ p} {q ⊗ r} {ι m} a b = 
    let  a′ = nest $ tile a                    ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ tileᵣ b                   ∶ Y ⟪ ι m ⊗ r ⟫ ⟪ q ⟫
         c′ = λ (i : P s) → sum (zipWith _⊞_) (K ε) 
                                λ k → mmt (a′ [ i ⊗ k ]) (b′ [ k ])
    in untileₗ $ unnest c′

  mmt {s = s ⊗ p} {q ⊗ r} {u ⊗ w} a b =
    let  a′ = nest $ tile a           ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ tile b           ∶ Y ⟪ w ⊗ r ⟫ ⟪ u ⊗ q ⟫
         c′ = λ { (i ⊗ j) → sum (zipWith _⊞_) (K ε) 
                            λ k → mmt (a′ [ i ⊗ k ]) 
                                      (b′ [ j ⊗ k ]) }
    in untile $ unnest c′



  -- This proof relies on the ability to commute sum:
  sum-commutes : (a : Z ⟪ p ⟫ ⟪ s ⟫)
               → sum (zipWith _⊞_) (K ε) a
                 ≈ₐ λ i → sum _⊞_ ε λ j → a [ j ] [ i ]
  sum-commutes a = sum-comm {a = a}


  blockedt-ok : (a : X ⟪ s ⊗ p ⟫) 
              → (b : Y ⟪ q ⊗ p ⟫)
              → mmt a b ≈ₐ mmt-canon a b
  blockedt-ok {s = ι m} {ι n} {ι k} a b 
    = refl-≈ₐ

  blockedt-ok {s = ι m} {ι n} {p ⊗ q} a b (i′ ⊗ (j′ ⊗ k′))
    = blockedt-ok a (tileₗ b ∘ (j′ ⊗_)) (i′ ⊗ k′)

  blockedt-ok {s = ι m} {p ⊗ q} {ι n} a b (i′ ⊗ j′)
    = let
        t k = mmt (tileᵣ a ∘ (k ⊗_)) (tileᵣ b ∘ (k ⊗_))
        r k = blockedt-ok (tileᵣ a ∘ (k ⊗_)) (tileᵣ b ∘ (k ⊗_)) (i′ ⊗ j′)
      in trans (sum-comm {a = t} (i′ ⊗ j′)) (sum-cong r)

  blockedt-ok {s = ι m} {p ⊗ q} {r ⊗ w} a b (i′ ⊗ (j′ ⊗ k′))
    = let
        t k = mmt (tileᵣ a ∘ (k ⊗_)) (tile b ∘ (j′ ⊗ k ⊗_))
        r k = blockedt-ok (tileᵣ a ∘ (k ⊗_)) (tile b ∘ (j′ ⊗ k ⊗_)) (i′ ⊗ k′)
      in trans (sum-comm {a = t} (i′ ⊗ k′)) (sum-cong r)

  blockedt-ok {s = s ⊗ p} {ι m} {ι n} a b (i′ ⊗ j′ ⊗ k′)
    = blockedt-ok (tileₗ a ∘ (i′ ⊗_)) b (j′ ⊗ k′)

  blockedt-ok {s = s ⊗ p} {ι m} {q ⊗ r} a b (i′ ⊗ j′ ⊗ (k′ ⊗ l′))
    = blockedt-ok (tileₗ a ∘ (i′ ⊗_)) (tileₗ b ∘ (k′ ⊗_)) (j′ ⊗ l′)

  blockedt-ok {s = s ⊗ p} {q ⊗ r} {ι m} a b (i′ ⊗ j′ ⊗ k′)
    = let
        t k = mmt (tile a ∘ (i′ ⊗ k ⊗_)) (tileᵣ b ∘ (k ⊗_))
        r k = blockedt-ok (tile a ∘ (i′ ⊗ k ⊗_)) (tileᵣ b ∘ (k ⊗_)) (j′ ⊗ k′)
      in trans (sum-comm {a = t} (j′ ⊗ k′)) (sum-cong r)

  blockedt-ok {s = s ⊗ p} {q ⊗ r} {u ⊗ w} a b (i′ ⊗ j′ ⊗ (k′ ⊗ l′))
    = let
        t k = mmt (tile a ∘ (i′ ⊗ k ⊗_)) (tile b ∘ (k′ ⊗ k ⊗_))
        r k = blockedt-ok (tile a ∘ (i′ ⊗ k ⊗_)) 
                          (tile b ∘ (k′ ⊗ k ⊗_)) (j′ ⊗ l′)
      in trans (sum-comm {a = t} (j′ ⊗ l′)) (sum-cong r)




  data R : S → S → S → Set where
    ι : R (ι m ⊗ ι n) (ι n ⊗ ι k) (ι m ⊗ ι k)
    σ : R s p q → R ((ι m ⊗ ι n) ⊗ s) ((ι n ⊗ ι k) ⊗ p) ((ι m ⊗ ι k) ⊗ q)

  mmx : R s p q → X ⟪ s ⟫ → Y ⟪ p ⟫ → Z ⟪ q ⟫
  mmx ι a b = mm-canon a b
  mmx (σ {s = s}{p}{q}{m = m}{n}{k} r) a b = let
    a′ = nest a
    b′ = nest b
    c′ : Z ⟪ q ⟫ ⟪ ι m ⊗ ι k ⟫ 
    c′ = λ { (i ⊗ j) → sum (zipWith _⊞_) (K ε) λ k′ → mmx r (a′ [ i ⊗ k′ ]) (b′ [ k′ ⊗ j ]) }
    in unnest c′

  
  record C (s p q : S) : Set₁ where
    constructor conv
    field
      ss pp qq : S
      s⇔ss×pp : Reshape s (ss ⊗ pp)
      p⇔pp×qq : Reshape p (pp ⊗ qq)
      ss×qq⇒q : Reshape (ss ⊗ qq) q

  toconv : R s p q → C s p q
  toconv {s = (ι m ⊗ ι n)} {(ι n ⊗ ι k)} {(ι m ⊗ ι k)} ι =
    conv (ι m)  (ι n) (ι k) eq eq eq
  toconv {s = (ι m ⊗ ι n ⊗ s)} {(ι n ⊗ ι k ⊗ p)} {(ι m ⊗ ι k ⊗ q)} (σ r) = let
    conv s′ p′ q′ s⇔s′p′ p⇔p′q′ s′q′⇔q  = toconv r
    in conv (ι m ⊗ s′) (ι n ⊗ p′) (ι k ⊗ q′)
            (rtile ∙ (eq ⊕ s⇔s′p′))
            (rtile ∙ (eq ⊕ p⇔p′q′))
            ((eq ⊕ s′q′⇔q) ∙ rtile)

  mmthm : (r : R s p q) (a : X ⟪ s ⟫) (b : Y ⟪ p ⟫)
        → let conv s′ p′ q′ fromA fromB toC = toconv r
          in mmx r a b ≈ₐ reshape toC (mm (reshape fromA a) (reshape fromB b))
  mmthm ι a b = refl-≈ₐ
  mmthm (σ r) a b (i′ ⊗ j′ ⊗ k′) with (k′ ⟨ C.ss×qq⇒q (toconv r) ⟩) 
                                      | inspect (k′ ⟨_⟩) (C.ss×qq⇒q (toconv r))
  ... | (k₁ ⊗ k₂) | Eq.[ pp ] rewrite sym pp = let
      A = (λ k′ → mmx r (λ j → a (i′ ⊗ k′ ⊗ j)) (λ j → b (k′ ⊗ j′ ⊗ j)))
      B = (λ k →
               mm
               (λ j →
                  a (i′ ⊗ k ⊗ (j ⟨ C.s⇔ss×pp (toconv r) ⟩)))
               (λ j →
                  b (k ⊗ j′ ⊗ (j ⟨ C.p⇔pp×qq (toconv r) ⟩))))

    in trans (sum₁-comm {a = A} {f = _⊞_}{e = ε} k′)
     $ trans (sum₁-cong {f = _⊞_}{e = ε} 
                        (λ j → mmthm r (λ j₁ → a (i′ ⊗ j ⊗ j₁)) 
                                      (λ j₁ → b (j ⊗ j′ ⊗ j₁)) k′))
       (sym (trans (sum₁-comm {a = B} {f = _⊞_}{e = ε} _) refl))


